import java.awt.*;

//2D point
class TwoDPoint {
	double y,z;

	TwoDPoint(double y,double z) {
		this.y = y;
		this.z = z;
	}
	
	//convert raw pixel values to internal coordinates
	TwoDPoint(int x,int y) {
		final double scaleFactor = .002; //scale factor determined empirically
		this.y = -scaleFactor*y;
		this.z = -scaleFactor*x;
	}
	
	//convert internal coordinates to pixel values and add to Polygon
	//scale factor is determined empirically
	void place(Polygon p) {
		final int scaleFactor = 320; //scale factor determined empirically
		p.addPoint((int)(scaleFactor*y),(int)(scaleFactor*z));
	}
}

//3D point
class ThreeDPoint {
	double x,y,z;

	ThreeDPoint(double x,double y,double z) {
		this.x = x;
		this.y = y;
		this.z = z;
	}
	
	//add points
	ThreeDPoint plus(ThreeDPoint p) {
		return new ThreeDPoint(x+p.x,y+p.y,z+p.z);
	}
	
	//subtract
	ThreeDPoint minus(ThreeDPoint p) {
		return new ThreeDPoint(x-p.x,y-p.y,z-p.z);
	}
	
	//inner product
	double dotProduct(ThreeDPoint p) {
		return x*p.x + y*p.y + z*p.z;
	}
	
	//Euclidean length of a vector
	double length() {
		return Math.sqrt(dotProduct(this));
	}
	
	//cosine of angle between 2 vectors
	double cos(ThreeDPoint p) {
		return dotProduct(p)/(length()*p.length());
	}
	
	//orient and project 3D point onto 2D viewing surface
	TwoDPoint project(Matrix orientation,ThreeDPoint viewerPosition) {
		return orientation.transpose().mult(this).project(viewerPosition);
	}
	
	//project 3D point onto 2D viewing surface
	TwoDPoint project(ThreeDPoint viewerPosition) {
		final double retinalDistance = 1; //determines perspective
		double alpha = retinalDistance/(viewerPosition.x-x);
		return new TwoDPoint(y*alpha,z*alpha);
	}
	
	//test whether a point is in field of vision
	boolean visible(ThreeDPoint viewerPosition) {
		final ThreeDPoint forward = new ThreeDPoint(-.05,0,0);
		return forward.dotProduct(minus(viewerPosition.plus(forward))) >= 0;
	}
}

//3D orthonormal matrix represented in row major order
class Matrix {
	ThreeDPoint x; //the first row
	ThreeDPoint y;
	ThreeDPoint z; //the last row

	Matrix(ThreeDPoint x, ThreeDPoint y, ThreeDPoint z) {
		this.x = x;
		this.y = y;
		this.z = z;
	}
	
	Matrix() { //rows not given--initialize to the identity matrix
		x = new ThreeDPoint(1,0,0);
		y = new ThreeDPoint(0,1,0);
		z = new ThreeDPoint(0,0,1);
	}
	
	//matrix-vector multiplication
	ThreeDPoint mult(ThreeDPoint p) {
		return new ThreeDPoint(x.dotProduct(p),y.dotProduct(p),z.dotProduct(p));
	}
	
	//matrix-matrix multiplication	
	Matrix mult(Matrix m) {
		Matrix mT = m.transpose();
		return new Matrix(mult(mT.x),mult(mT.y),mult(mT.z)).transpose();
	}
	
	//transpose of a matrix
	//for orthonormal matrices, this is the inverse
	Matrix transpose() {
		ThreeDPoint a = new ThreeDPoint(x.x,y.x,z.x);
		ThreeDPoint b = new ThreeDPoint(x.y,y.y,z.y);
		ThreeDPoint c = new ThreeDPoint(x.z,y.z,z.z);
		return new Matrix(a,b,c);
	}
}

//a square is a flat polygon in our figure to be rendered
class Square {
	ThreeDPoint center; //a point in the square
	ThreeDPoint normal; //the normal vector
	Color color;
	//template for corners of squares
	//these will be translated when a new square is created
	static final ThreeDPoint[][] baseCorners = {
		{
			new ThreeDPoint(0,1,1),
			new ThreeDPoint(0,-1,1),
			new ThreeDPoint(0,-1,-1),
			new ThreeDPoint(0,1,-1)
		}, {
			new ThreeDPoint(1,0,1),
			new ThreeDPoint(-1,0,1),
			new ThreeDPoint(-1,0,-1),
			new ThreeDPoint(1,0,-1)
		}, {
			new ThreeDPoint(1,1,0),
			new ThreeDPoint(-1,1,0),
			new ThreeDPoint(-1,-1,0),
			new ThreeDPoint(1,-1,0)
		}
	};
	//corners of the square
	ThreeDPoint[] corners = new ThreeDPoint[4];
	ThreeDPoint[] orientedCorners = new ThreeDPoint[4];
	ThreeDPoint[] truncatedCorners = new ThreeDPoint[5];
	
	//create a new square by translating the corner template	
	Square(BSPTree t, ThreeDPoint center, ThreeDPoint normal, Color color) {
		this.center = center;
		this.normal = normal;
		this.color = color;
		int cornersIndex;
		if (normal.x != 0) cornersIndex = 0;
		else if (normal.y != 0) cornersIndex = 1;
		else cornersIndex = 2;
		for (int i = 0; i < 4; i++) {
			corners[i] = center.plus(baseCorners[cornersIndex][i]);
		}
		t.insert(this); //insert in BSPTree
	}
	
	//project the square onto 2D for rendering
	//calculate proper orientation first
	Polygon project(Matrix orientation, ThreeDPoint viewerPosition) {
		Polygon p = new Polygon(); //polygon object to be created
		TwoDPoint corner; //2D projection of a corner
		for (int i = 0; i < 4; i++) { //orient corners
			orientedCorners[i] = orientation.transpose().mult(corners[i]);
		}		
		for (int i = 0; i < 4; i++) {
			if (orientedCorners[i].visible(viewerPosition)) {
				orientedCorners[i].project(viewerPosition).place(p);
			} else { //not visible--truncate Square by viewer plane
				if (orientedCorners[(i+3)%4].visible(viewerPosition)) {
					ThreeDPoint s = truncate(orientedCorners[(i+3)%4],orientedCorners[i],viewerPosition);
					s.project(viewerPosition).place(p);
				}
				if (orientedCorners[(i+1)%4].visible(viewerPosition)) {
					ThreeDPoint s = truncate(orientedCorners[(i+1)%4],orientedCorners[i],viewerPosition);
					s.project(viewerPosition).place(p);
				}
			}			
		}
		return p;
	}
	
	//find intersection with viewer's field of vision	
	ThreeDPoint truncate(ThreeDPoint s, ThreeDPoint t, ThreeDPoint vp) {
		final ThreeDPoint forward = new ThreeDPoint(-.05,0,0);
		vp = vp.plus(forward);
		double a = vp.minus(t).dotProduct(vp);
		double b = s.minus(t).dotProduct(vp);
		double alpha = (b == 0)?0:(a/b);
		double u = alpha*s.x + (1 - alpha)*t.x;
		double v = alpha*s.y + (1 - alpha)*t.y;
		double w = alpha*s.z + (1 - alpha)*t.z;
		return new ThreeDPoint(u,v,w);		
	}
	
	//is this square behind the given point?
	boolean behind(ThreeDPoint p) {
		return p.minus(center).dotProduct(normal) >= 0;
	}
}

//a cube has 6 squares
class Cube {
	ThreeDPoint center;
	Color color;
	Square faces[] = new Square[6];
	static final ThreeDPoint[] normals = {
		new ThreeDPoint(1,0,0),
		new ThreeDPoint(-1,0,0),
		new ThreeDPoint(0,1,0),
		new ThreeDPoint(0,-1,0),
		new ThreeDPoint(0,0,1),
		new ThreeDPoint(0,0,-1)
	};		
	
	Cube(BSPTree t, ThreeDPoint center, Color color) {
		this.center = center;
		this.color = color;
		for (int i = 0; i < 6; i++) {
			faces[i] = new Square(t,center.plus(normals[i]),normals[i],color);
		}
	}
	
	//is the given point inside this cube?
	boolean inside(ThreeDPoint p) {
		for (int i = 0; i < 6; i++) {
			if (faces[i].behind(p)) return false;
		}
		return true;
	}
}

//8 cubes of different colors
class Cubes {
	BSPTree t = new BSPTree(); //keep the faces in a BSP tree
	Cube[] cube = new Cube[8];
	
	Cubes() {
		cube[0] = new Cube(t,new ThreeDPoint(2,2,2),Color.red);
		cube[1] = new Cube(t,new ThreeDPoint(2,2,-2),Color.yellow);
		cube[2] = new Cube(t,new ThreeDPoint(2,-2,2),Color.blue);
		cube[3] = new Cube(t,new ThreeDPoint(-2,2,2),Color.green);
		cube[4] = new Cube(t,new ThreeDPoint(2,-2,-2),Color.cyan);
		cube[5] = new Cube(t,new ThreeDPoint(-2,2,-2),Color.magenta);
		cube[6] = new Cube(t,new ThreeDPoint(-2,-2,2),Color.orange);
		cube[7] = new Cube(t,new ThreeDPoint(-2,-2,-2),Color.pink);
	}
	
	//render the Squares
	void render(Figure f, ThreeDPoint viewpoint) {
		t.enumerate(f,viewpoint); //walk the BSP tree
	}
	
	//is the given point inside any of the cubes?
	boolean inside(ThreeDPoint p) {
		for (int i = 0; i < 8; i++) {
			if (cube[i].inside(p)) return true;
		}
		return false;
	}

}


